#include "mscaling.hpp"

namespace Mscaling{

  //
  // distance between two points
  //
  Distance distance(const Point& x, const Point& y)
  {
    Distance total = 0.0;
    Distance diff;
    
    Point::const_iterator cpx=x.begin(); 
    Point::const_iterator cpy=y.begin();
    Point::const_iterator cpx_end=x.end();
    for(;cpx!=cpx_end;++cpx,++cpy){
      diff = *cpx - *cpy;
      total += (diff * diff); 
    }
    return total;  // no need to take sqrt, which is monotonic
  };

  // compute the distance among all the points
  //  
  void all_distances(const Points& ps, DistMatrix & dist){
    
    size_t size = ps.size();
    
    for (unsigned int i=0; i < size; i++)
      for (unsigned int j=0; j < size; j++){
	if (i==j)
	  continue;
	dist(i,j) = distance(ps[i], ps[j]);
      }
    //	std::cerr << "m(" << i << "," << j << ")=" << dist(i,j) << " ";
  }

  
  // Ramdom generate points
  void random_gen(Points & ps)
  {

    Points::iterator p_i, p_end;
    p_i = ps.begin(); p_end = ps.end();
    for (;p_i!=p_end; p_i++){   // each point
      
      (*p_i)[0]= rand() % 1000 + 1; // just some random assignment
      (*p_i)[1]= rand() % 1000 + 1; // just some random assignment
    }
  };


  //
  // scale points down
  void scaledown(Points ps, double rate, 
		 unsigned int num_iterations)
  {
    using namespace boost::numeric::ublas;

    size_t size = ps.size();
    matrix<double> real_dist(size, size);  // real distance
    matrix<double> fake_dist(size, size);  // fake distance
    double total_error, error_term, last_error;


    Points ps_random(size); // create a random allocation of points
    random_gen(ps_random);

    Points gradient(size);
 
    // compute real distance among real points
    //
    all_distances(ps, real_dist);  

    // for all iteration
    //
    for (unsigned int ni = 0; ni < num_iterations; ni++){

      // distance 
      // between real and fake points
      for (unsigned int i = 0 ; i < size; i++)
	for (unsigned int j = 0; j < size; j++)
	  fake_dist(i, j) = distance(ps[i], ps_random[j]);

      // init gradient and error
      for (unsigned int i = 0 ; i < size; i++)
	gradient[i][0] = gradient[i][1] = 0.0;
      total_error = 0;

      // computer grandient and error_term
      // 
      for (unsigned int i = 0 ; i < size; i++)
	for (unsigned int j = 0; j < size; j++){
	  if (i==j)
	    continue;

	  error_term=(fake_dist(i,j) - real_dist(i,j))/fake_dist(i,j);

	  //std::cerr << "error_term=" << error_term << std::endl;

	  gradient[i][0]+=
	    ((ps_random[i][0]-ps_random[j][0])/fake_dist(i,j))*error_term;
	  gradient[i][1]+=
	    ((ps_random[i][1]-ps_random[j][1])/fake_dist(i,j))*error_term;

	  total_error += abs(error_term);
	}
      
      // have we reached the optimum
      // 
      if (ni > 1 && last_error <= total_error) {

      std::cout << "it=" << ni << " last_error=" << last_error 
		<< " total_error=" << total_error << std::endl;
	break;
      }

      last_error = total_error;

      // try to move the points
      //
      for (unsigned int i=0; i < size; i++){
	ps_random[i][0] -= rate*gradient[i][0];
	ps_random[i][1] -= rate*gradient[i][1];
      }
    }// num_iterations
  };

  // Dump collection of Points
  //
  std::ostream& operator << (std::ostream& os, Points& cps){
    BOOST_FOREACH(Points::value_type p, cps){
      os<<p[0] << ' ' << p[1]<<std::endl;
    }
    return os;
  }


};
